/*------------------------------------------------------------------------------------------------

myPing.C

C program to notify us when Internet is restored.  Useful when our ISP goes down. 
Program will beep when Internet is restored so we don't have to keep checking manually.

Target OS = MS Windows -- most any version 1998 and later should be fine.
Calls MS ping utility -- reads output from ping to discern internet status.

Author Greg Spears
Given to the Public domain
2021.12.09

Don't be a lamer.  If you use my code, greet me!  

*--------------------------------------------------------------------------------------------------*/

#define WIN32_LEAN_AND_MEAN
#undef UNICODE
#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <stdarg.h>
#include <conio.h>
#include <windows.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <process.h>
#include <stdlib.h>

#pragma warning( disable : 4996) //Silence warn about new strcpy is safer and such

int screen_w, screen_h, screen_buffer_w, screen_buffer_h, screencolor;
HANDLE hStdIn, hStdOut; /* standard input, output handles */
CHAR *szScreenBuffer; /* buffer to read the index number into */

enum{
INTERNET_STATUS_FALSE,
INTERNET_STATUS_TRUE,
INTERNET_STATUS_UNKNOWN};




/*---------------------------------------------------------------------------------
my_printf()

Adds color to my printed output.  Restores screencolor after print.

*-----------------------------------------------------------------------------------*/
void my_printf(char *msg, ... )
{
va_list  argptr;          /* Argument list pointer    */
char szBuff[5000];         /* Buffer to build string into   */

    va_start( argptr, msg );      /* Initialize va_ functions */
    vsprintf( szBuff, msg, argptr );   /* prints string to buffer  */
    va_end( argptr );         /* Close va_ functions      */

	/* Get screencolor (already done at start up ) See:  ConsoleCoordsInit() */
	//screencolor = GetConsoleTextAttribute(hStdOut );

#define BACKGROUND_WHITE  (WORD) 0x00f0
#define BACKGROUND_CYAN   (WORD) 0x0030
#define FOREGROUND_YELLOW (WORD) 0x0006
#define FOREGROUND_CYAN   (WORD) 0x0003
#define FOREGROUND_WHITE  (WORD) 0x0007

	/* Change foreground white */
	SetConsoleTextAttribute(hStdOut, FOREGROUND_WHITE );

	/* print mssg */
	printf(szBuff);


	/* Restore screen color from step 1 */
	SetConsoleTextAttribute(hStdOut, screencolor );
}

/*-------------------------------------------------------------------------------

myDelay()

Checks if keyboard hit or any console event during delay.
Breaks early if event sensed -- ie: user hitting keybord ESC for exit

NOTE: clock() is used because multiple calls to Sleep(1) result in much 
     more delay than (NumCalls * 1ms)
     -- maybe 10X more --- 2000 calls of Sleep(1) > 10 secs (expected 2 secs)

*--------------------------------------------------------------------------------*/
void myDelay(DWORD ms)
{
DWORD dwInputEvents=0;
DWORD count = 0, start = clock();

	//my_printf("\nDELAY: %dms\n\n", ms);

	while(clock() - start < ms)
	{
		Sleep(1);
		count++;
		GetNumberOfConsoleInputEvents(hStdIn, &dwInputEvents);
		if( dwInputEvents )
		{
			//my_printf("Breaking for console events %d\n", dwInputEvents);
			break;
		}
	}
	//my_printf("Exiting delay loop after %d iterations\n", count);// typically 120 iterations for 2000ms.
	
}


/*---------------------------------------------------------------------------------
* FUNCTION: ConsoleCoordsInit(HANDLE hCon)                          
*                                                                   
* PURPROSE: Fill global screen variables with data                  
*                                                                   
* INPUT: the handle to get the information for                      
*                                                                   
* RETURNS: TRUE /FALSE                                              
---------------------------------------------------------------------------------*/
SHORT ConsoleCoordsInit(HANDLE hCon)
{
  CONSOLE_SCREEN_BUFFER_INFO csbi;
  BOOL bSuccess;

  bSuccess = GetConsoleScreenBufferInfo(hCon, &csbi);

	screen_w = ( csbi.srWindow.Right -csbi.srWindow.Left+1);
	screen_h = ( csbi.srWindow.Bottom -csbi.srWindow.Top+1);
	screen_buffer_w = csbi.dwSize.X;
	screen_buffer_h = csbi.dwSize.Y;
	screencolor = csbi.wAttributes;

  return(bSuccess);
}


/*------------------------------------------------------------------------------------

  char *strnstr()

   returns pointer to first occurrance of substring in string
    NULL if not found

 params:

      str - string to search
      pattern - string to search for
	  max -- max number of chars to search in str

*-----------------------------------------------------------------------------------*/
char *strnstr(const char *str, const char *pattern, int max)
{
int s,p,p1, offs1=0, patlen = strlen(pattern);
const char *pat;

    p1 = (*pattern);

    do {
        // Speed optimize: scan string for a match on 1st character
		while(s = str[offs1++])
        {
            if(s == p1)
            {
                break;
            }
        }
		/* Halt if search will exceed user-defined search depth */
		if (offs1 + patlen > max+1 )
		{
			//puts("safety break");
			break;
		}

        if(s)
        {
			int offs2=0;
            // Scan for continued match
            pat=&pattern[1];  //we know 1st char, start at 2nd
            do{
                if(!*pat)  //have we reached the end of search string?
                    return (char *)&str[offs1-1]; //Y: and we've been thru whole string - a match
                s = str[offs1 + offs2++];
                p = (*pat++);
            }while(s==p);
        }

    }while(s);

    return(NULL);
}


/*---------------------------------------------------------------------------------

scan_buffer

Scan and search screen buffer to discern ping status

Looks for ping output info.  (We will call ping with no options)
* WITHOUT -t option -- sample responses:
- "Ping request could not find host google.com. Please check the name and try again."  -- No internet
- "Reply from 142.251.40.78: bytes=32 time=26ms TTL=111"   -- Postive Internet 
- "General failure."  -- Cable to PC unplugged /  
- "Request timed out."  -- Network not established -- router booting
- "PING: transmit failed. General failure."  -- Network not established, router booting

WITH -t option
* Reply from 10.0.0.1: Destination net unreachable." -- cable to router  unplugged 

Micrsoft screen buffer info:
- The entire buffer is padded with spaces (char 32 ) on startup
- Microsoft fills the buffer with screen data from the zero offset. -- (not from the bottom )
- When the buffer is full, I assume the entire contents are shifted up:
  - New line is last from screen, 1st line is lost, old 2nd line is now 1st, etc...


Algorithm:
- Go to last line of buffer.  Always eval and report most current line -  priority!!
- compare contents to the Key ping outputs:
	- Reply from [...] bytes=32 time=26ms TTL=111"
	- Ping request could not find host [...]. Please check the name and try again."
	- General failure
	- timeout 
- If no match, go to the previous line, continue from compare step

---------------------------------------------------------------------------------*/
int scan_buffer(char *buff, int buffw, int buffh)
{
int offset = (buffw * (buffh-1)); /* offset placed at start of last line */
char *sPtr;

	if(!buff)
		return 0;


	/* Buffer is filled from top down.  Scan backward aand find 1st non-whitespace at bottom of buffer.*/
	while(offset > -1 )
	{
		sPtr = &buff[offset];

		if(*sPtr != 32 )
		{
			/* String compare to known messages */
			if(strnstr(sPtr, "Destination net unreachable", buffw) )
			{
				my_printf("\nNo Internet.\n");
				return INTERNET_STATUS_FALSE;
			}
			if(strnstr(sPtr, "Reply from", buffw) )
			{
				my_printf("\nWe have Internet!\n");
				return INTERNET_STATUS_TRUE;
			}
			if(strnstr(sPtr, "General failure", buffw) )
			{
				my_printf("\nNo Internet.\n");
				my_printf("Cable unplugged/no network\n");
				return INTERNET_STATUS_FALSE;
			}

			if(strnstr(sPtr, "Ping request could not find host", buffw) )
			{
				my_printf("\nNo Internet.\n");
				return INTERNET_STATUS_FALSE;
			}

			if(strnstr(sPtr, "Request timed out.", buffw) )
			{
				my_printf("Time Out sensed:  Internet status unknown\n");
				return INTERNET_STATUS_UNKNOWN;
			}
		}
		offset -= buffw; //goto previous line
	}

	return FALSE;
}




/*---------------------------------------------------------------------------------

ck_console_events()

---------------------------------------------------------------------------------*/
ck_console_events()
{
BOOL bSuccess;
DWORD dwInputEvents;
INPUT_RECORD inputBuffer;


 /* read an input events from the input event queue */
	GetNumberOfConsoleInputEvents(hStdIn, &dwInputEvents);
	if( dwInputEvents )
	{
		bSuccess = ReadConsoleInput(hStdIn, &inputBuffer, 1, &dwInputEvents);
		switch (inputBuffer.EventType)
		{
			case KEY_EVENT:
			if (inputBuffer.Event.KeyEvent.bKeyDown)
			{
			  /* Allow exit on ESCAPE key */
				if( inputBuffer.Event.KeyEvent.wVirtualKeyCode == VK_ESCAPE )
				{
					exit(0);
				}
			}
			break;
			
			case MOUSE_EVENT:  // not interested for now
			break;

			case WINDOW_BUFFER_SIZE_EVENT:   //get new screen coords
				ConsoleCoordsInit(hStdOut);
				if ( szScreenBuffer ) 
					free(szScreenBuffer);
				szScreenBuffer = (char *) malloc(screen_buffer_w * screen_buffer_h);
				my_printf("window: %d x %d\nWindow Buffer %d x %d ", 
					screen_w, screen_h, screen_buffer_w, screen_buffer_h);
				if(szScreenBuffer )
					my_printf("Mem alloc success\n");
				else
				{
					my_printf("Mem alloc for screen resize failed...exiting.\n");
					_getch();
					exit(0);
				}
			break;
		  } /* switch */

		FlushConsoleInputBuffer(hStdIn);
	}

	return TRUE;
}



/*---------------------------------------------------------------------------------

main()

---------------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
time_t rawtime;
struct tm timeinfo;
int errn;
BOOL bSuccess;
/* array of console input event records */
CHAR bOutBuf[256]; /* buffer to format event information into */
/* used to get the new console screen buffer size if it is changed */
COORD coordScreen; /* used when reading the index number from the screen */
DWORD cCharsRead;
int  current_internet_status=-1, prev_internet_status = -1, startup = TRUE, just_regained_internet = FALSE;


	hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	hStdIn = GetStdHandle(STD_INPUT_HANDLE);

	ConsoleCoordsInit(hStdOut);
	
	szScreenBuffer = (char *) malloc(screen_buffer_w * screen_buffer_h);

	if(! szScreenBuffer )
	{
		my_printf("Mem alloc for screen buffer failed.... exiting...\n");
		_getch();
		return 0;
	}
	

	while(1)
	{
		time(&rawtime);
		errn = localtime_s(&timeinfo, &rawtime);
		strftime (bOutBuf,80,"%Y.%m.%d - %H:%M:%S", &timeinfo);
		my_printf("\n%s\n",bOutBuf);

		//my_printf("ping google.com");
		system("ping google.com");

		ck_console_events();

		if( ! szScreenBuffer )
			continue;

		coordScreen.X = coordScreen.Y = 0;
		//my_printf("Screenbuffer size: %d \n", screen_buffer_w * screen_buffer_h );
		bSuccess = ReadConsoleOutputCharacter(hStdOut, szScreenBuffer, (screen_buffer_w * screen_buffer_h), coordScreen, &cCharsRead);
		
		current_internet_status  = scan_buffer(szScreenBuffer, screen_buffer_w, screen_buffer_h);

		if( current_internet_status !=  prev_internet_status && !startup )
		{
			/* Status changed and it isn't due to program startup */
			switch (current_internet_status)
			{
				case INTERNET_STATUS_TRUE:
					if(prev_internet_status == INTERNET_STATUS_FALSE)
					{
						/*We have internet and this is a new event -- beep alot*/
						just_regained_internet = TRUE;
					}
				break;

				case INTERNET_STATUS_FALSE:
					/*We have just lost internet  -- beep once long*/
					Beep(880, 1000);
				break;
				
				case INTERNET_STATUS_UNKNOWN:
					current_internet_status =  prev_internet_status;
				break;
			}
		}

		startup = FALSE;
		prev_internet_status = current_internet_status;

		if( current_internet_status == INTERNET_STATUS_FALSE )
				just_regained_internet = FALSE;
	
		if ( just_regained_internet && current_internet_status != INTERNET_STATUS_UNKNOWN)
		{
			Beep(1760, 500);  //1760 is a very audible frequency -- 2 octaves above concert A 
			Beep(1760, 500);
			Beep(1760, 500);
		}

		ck_console_events();

		my_printf("ESC key or CTRL+C to quit.\n");
		myDelay(2000);

		ck_console_events();
	}

	return 0;
}
